searchNodes=[{"doc":"ExRaft provides an Elixir implementation and an easy to use API for the raft consensus protocol. Internals Servers When starting an ExRaft server, there are 2 processes spawned: a Supervisor process that ensures that the logs are not lost when the server crashes the actual server process Note that the Supervisor will generate an additional atom to register itself (for more information, check out the ExRaft.start_server/3 function). Timeouts When calling ExRaft.write/3 , ExRaft.add_server/3 or ExRaft.remove_server/3 , you may experience timeouts. This is most likely caused by the majority of the cluster not being available to the leader. Note that even though the call may time out, once the majority comes back online your command might still be applied, provided the rest of the cluster did not advance their log in the meantime. Cluster membership changes Dynamic cluster membership is supported with one-at-a-time changes to the configuration, but as of now new servers are added without any sort of catch-up mechanism or validation (these features will be present in a future release). Servers can be added to- or removed from the cluster by making ExRaft.add_server/3 or ExRaft.remove_server/3 calls to the leader. Log compaction As of now log compaction is not supported but is planned to be implemented and released in the future. Example init_arg = [ very_cool : true ] initial_config = [ raft1 : node ( ) , raft2 : node ( ) ] # after starting these servers, they will time out and eventually elect a leader # amongst themselves { :ok , _ } = ExRaft . start_server ( YourStateMachine , init_arg , name : :raft1 , initial_config : initial_config ) { :ok , _ } = ExRaft . start_server ( YourStateMachine , init_arg , name : :raft2 , initial_config : initial_config ) # this server can never become leader unless its added to an existing cluster # since it doesn&#39;t know of any other server but is required to achieve # a minimum majority of 2 in elections and log replication { :ok , _ } = ExRaft . start_server ( YourStateMachine , init_arg , name : :raft3 , min_majority : 2 ) # we could pick any server to await the leader leader = ExRaft . await_leader ( :raft1 ) :ok = ExRaft . add_server ( leader , :raft3 ) # the success result of write/3 depends on the state machine :ok = ExRaft . write ( leader , :hello ) # making a write/3 call to a follower results in an error follower = List . delete ( initial_config , leader ) |&gt; List . first ( ) { :error , { :redirect , ^ leader } } = ExRaft . write ( follower , :hello ) # if we stop the active leader (or it crashes) # the rest of the cluster will be able to recover :ok = ExRaft . stop_server ( leader ) :ok = ExRaft . trigger_election ( :raft3 ) new_leader = ExRaft . await_leader ( :raft3 ) true = Enum . member? ( [ raft2 : node ( ) , raft3 : node ( ) ] , new_leader )","ref":"ExRaft.html","title":"ExRaft","type":"module"},{"doc":"Makes a synchronous call to server to add a new server to the configuration. Return values If all checks pass, returns :ok once the majority of the new configuration (including new_server ) knows of new_server being a member of the cluster. (Returns :ok immediately if new_server is already a member of the cluster.) Otherwise returns {:error, reason} , where reason is one of the following: :no_commit - server has yet to commit an entry in it's term as leader. :unstable_config - another configuration change is pending. can also return any of the error reasons specified at write/3","ref":"ExRaft.html#add_server/3","title":"ExRaft.add_server/3","type":"function"},{"doc":"Makes a synchronous call to server and waits until server learns of; or becomes the cluster leader. Returns the elected cluster leader.","ref":"ExRaft.html#await_leader/2","title":"ExRaft.await_leader/2","type":"function"},{"doc":"Sends an asynchronous request to server to inspect and write its context or log to the device.","ref":"ExRaft.html#inspect/2","title":"ExRaft.inspect/2","type":"function"},{"doc":"Makes a synchronous call to server to retrieve its known leader and waits for its reply.","ref":"ExRaft.html#leader/2","title":"ExRaft.leader/2","type":"function"},{"doc":"Makes a synchronous call to server to read query from the state and waits for its reply. Return values If server is leader, it returns the result of calling the handle_read/2 function of the underlying state machine (see ExRaft.StateMachine ). Otherwise if server is not leader, it returns the same error as write/3 .","ref":"ExRaft.html#read/3","title":"ExRaft.read/3","type":"function"},{"doc":"Same as read/3 , but if :dirty_read is set to true (see start_server/3 ), followers are also allowed to execute the query. If :dirty_read was set to false , server never replies.","ref":"ExRaft.html#read_dirty/3","title":"ExRaft.read_dirty/3","type":"function"},{"doc":"Makes a synchronous call to server to remove a server from the configuration. Return values If all checks pass, returns :ok once the majority of the new configuration (without server_to_remove ) knows of server_to_remove not being a member of the cluster. (Returns :ok immediately if server_to_remove is not a member of the cluster.) Otherwise returns {:error, reason} , where reason is one of the following: :last_server - server_to_remove is the last server in the configuration. :no_commit - server has yet to commit an entry in it's term as leader. :unstable_config - another configuration change is pending. can also return any of the error reasons specified at write/3","ref":"ExRaft.html#remove_server/3","title":"ExRaft.remove_server/3","type":"function"},{"doc":"Same as start_server/3 but sets [] as init_arg .","ref":"ExRaft.html#start_server/2","title":"ExRaft.start_server/2","type":"function"},{"doc":"Starts a server process linked to the current process. Once the server is started, the init/1 function (see ExRaft.StateMachine ) of the given module is called with init_arg as its argument to initialize the server. Options :name - used for name registration as described in the &quot;Name Registration&quot; section of the GenServer documentation. Note that only local names (atoms) are accepted. By design ExRaft will create an additional atom for the server supervisor under :&quot;Elixir.ExRaft.Supervisor.\#{name}&quot; . This is the only option required to ensure internal server communication. :debug - can be set to true to enable verbose server logs. Defaults to false . :dirty_read - can be set to false to disable reading &quot;dirty&quot; (maybe out of date) state from followers. Defaults to true . :initial_config - used to start the server with a known set of peers. Defaults to [] . :min_majority - used to set a minimum expected majority. When set, servers will decide on achieving majority in elections and log replication based on the active configuration and this value (the greater value will be used). As an example, with a value of 2 in a 2 member cluster, the required majority will be 2, with the same value in a 5 member cluster, the required majority will be 3. Defaults to nil . :min_election_timeout - used to tune the minimum amount of time in milliseconds, that must pass before a follower times out. Note that the actual timeout will be picked randomly between :min_election_timeout and :max_election_timeout . Defaults to 1_000 . :max_election_timeout - used to tune the maximum amount of time in milliseconds, that must pass before a follower times out. Note that the actual timeout will be picked randomly between :min_election_timeout and :max_election_timeout . Defaults to 10_000 . :heartbeat_timeout - used to set the heartbeat timeout (aka.: the time that will pass between leader heartbeats in log replication) in milliseconds. Defaults to 100 . :batch_size - used to set the number of entries sent by leaders to followers in heartbeat RPCs. Defaults to 100 .","ref":"ExRaft.html#start_server/3","title":"ExRaft.start_server/3","type":"function"},{"doc":"Makes a synchronous call to server to retrieve its state and waits for its reply.","ref":"ExRaft.html#state/2","title":"ExRaft.state/2","type":"function"},{"doc":"Synchronously stops the server with the given reason . server will call the terminate/2 function of the underlying state machine before exiting (see ExRaft.StateMachine ). This function keeps OTP semantics regarding error reporting. If the reason is any other than :normal , :shutdown or {:shutdown, _} , an error report is logged.","ref":"ExRaft.html#stop_server/3","title":"ExRaft.stop_server/3","type":"function"},{"doc":"Makes a synchronous call to server to start a new election. server replies with :ok immediately and becomes candidate to start a new election unless server is already a candidate in which case the reply will still be :ok but nothing happens.","ref":"ExRaft.html#trigger_election/2","title":"ExRaft.trigger_election/2","type":"function"},{"doc":"Makes a synchronous call to server to apply command to the state and waits for its reply. Return values If server is leader, it appends command to its log, and once the majority of the cluster replicates command , it applies command to the state by calling the handle_write/2 function of the underlying state machine (see ExRaft.StateMachine ). It then returns reply from the handle_write/2 call. Otherwise if server is not leader, it returns {:error, reason} where reason is one of the following: :no_leader - server does not know the leader of the cluster. {:redirect, leader} - where leader is the leader of the cluster.","ref":"ExRaft.html#write/3","title":"ExRaft.write/3","type":"function"},{"doc":"Option values used by the start_server functions. See start_server/3 for more information.","ref":"ExRaft.html#t:option/0","title":"ExRaft.option/0","type":"type"},{"doc":"Options used by the start_server functions. See start_server/3 for more information.","ref":"ExRaft.html#t:options/0","title":"ExRaft.options/0","type":"type"},{"doc":"Reference to a peer. Used internally by ExRaft servers to communicate with each other.","ref":"ExRaft.html#t:peer/0","title":"ExRaft.peer/0","type":"type"},{"doc":"Reference to a server. Used to make calls to ExRaft servers.","ref":"ExRaft.html#t:server/0","title":"ExRaft.server/0","type":"type"},{"doc":"A behaviour module for implementing consistently replicated state machines. Example To start things off, let's see a code example implementing a simple key value store in ExRaft style: defmodule KeyValueStore do @behaviour ExRaft.StateMachine @impl true def init ( init_state ) do { :ok , init_state } end @impl true def command? ( { :put , _ , _ } , _ ) , do : true def command? ( _ , _ ) , do : false @impl true def handle_write ( { :put , key , value } , state ) do new_state = Map . put ( state , key , value ) reply = { :ok , &quot;I GOT IT BOSS&quot; } { reply , new_state } end @impl true def handle_read ( key , state ) do { :reply , Map . get ( state , key ) } end end initial_config = [ foo : node ( ) , bar : node ( ) , baz : node ( ) ] init_state = %{ } { :ok , _ } = ExRaft . start_server ( KeyValueStore , init_state , name : :foo , initial_config : initial_config ) { :ok , _ } = ExRaft . start_server ( KeyValueStore , init_state , name : :bar , initial_config : initial_config ) { :ok , _ } = ExRaft . start_server ( KeyValueStore , init_state , name : :baz , initial_config : initial_config ) leader = ExRaft . await_leader ( :foo ) # leader is now one of [{:foo, node()}, {:bar, node()}, {:baz, node()}] ExRaft . read ( leader , :some_key ) # nil ExRaft . write ( KeyValueStore , { :put , :some_key , :some_value } ) # {:ok, &quot;I GOT IT BOSS&quot;} ExRaft . read ( leader , :some_key ) # :some_value We start a member of our KeyValueStore cluster by calling ExRaft.start_server/3 , passing the module with the state machine implementation and its initial state (in this example this is just an empty map). In the example above we start a 3 server cluster with the names :foo , :bar and :baz . Any of these servers may become the cluster leader at startup or when the old leader fails, that is why we first wait for a leader to be elected by calling ExRaft.await_leader/2 . After a leader was elected we can write to- and read from our replicated key value store by calling ExRaft.write/3 and ExRaft.read/3 respectively. Once the ExRaft.write/3 call returns with our state machine reply ( {:ok, &quot;I GOT IT BOSS&quot;} ) we know that the majority of our cluster (2 servers in this example) knows that :some_key is in fact :some_value . This means that if the leader crashes or a network split happens between the leader and the rest of the cluster, the remaining 2 servers will still be able to elect a new leader and this new leader will still know of :some_key being :some_value . In fact if we were to issue an ExRaft.read_dirty/3 call on the followers after writing :some_key to the state at least one, if not both of them would reply with :some_value . Improvements In the above example we interacted with the key value store using the ExRaft module. This is not ideal since we don't want our users to necessarily know how to use ExRaft . Also we started the servers by calling ExRaft.start_server/3 directly. It would be better if we started them as part of a supervision tree. So let's fix these issues: defmodule KeyValueStore do use ExRaft.StateMachine @init_state %{ } def start_link ( opts ) , do : ExRaft . start_server ( __MODULE__ , @init_state , opts ) def put ( server , key , value ) , do : ExRaft . write ( server , { :put , key , value } ) def get ( server , key ) , do : ExRaft . read ( server , key ) @impl true def init ( init_state ) do { :ok , init_state } end @impl true def command? ( { :put , _ , _ } , _ ) , do : true def command? ( _ , _ ) , do : false @impl true def handle_write ( { :put , key , value } , state ) do new_state = Map . put ( state , key , value ) reply = { :ok , &quot;I GOT IT BOSS&quot; } { reply , new_state } end @impl true def handle_read ( key , state ) do { :reply , Map . get ( state , key ) } end end Now we can simply call KeyValueStore.put/3 and KeyValueStore.get/3 to write to- and read from our replicated key value store. Supervision When we invoke use ExRaft.StateMachine , two things happen: It defines that the current module implements the ExRaft.StateMachine behaviour ( @behaviour ExRaft.StateMachine ). It also defines an overridable child_spec/1 function, that allows us to start the KeyValueStore directly under a supervisor. children = [ { KeyValueStore , name : :foo , initial_config : [ ... ] } ] Supervisor . start_link ( children , strategy : :one_for_one ) This will invoke KeyValueStore.start_link/1 , lucky for us we already defined this function in the improved implementation. Note Be aware that all callbacks (except for init/1 and terminate/2 ) block the server until they return, so it is recommended to try and keep any heavy lifting outside of them or adjusting the :min_election_timeout and :max_election_timeout options of the server (see ExRaft.start_server/3 ). Also note that setting a high election timeout range may cause the cluster to stay without leader and thus become unresponsive for a longer period of time.","ref":"ExRaft.StateMachine.html","title":"ExRaft.StateMachine","type":"behaviour"},{"doc":"Invoked to check commands upon ExRaft.write/3 calls. Returning true will cause the server to continue with log replication and eventually apply command to the state . Returning false will make the server ignore the command, causing the client to eventually time out. This callback is optional. If one is not implemented, all commands will be replicated and eventually applied to the state .","ref":"ExRaft.StateMachine.html#c:command?/2","title":"ExRaft.StateMachine.command?/2","type":"callback"},{"doc":"Invoked to handle queries from ExRaft.read/3 and when enabled ExRaft.read_dirty/3 calls. Returning {:reply, reply} sends the response reply to the caller and continues the loop. Returning :noreply does not send a response to the caller and continues the loop.","ref":"ExRaft.StateMachine.html#c:handle_read/2","title":"ExRaft.StateMachine.handle_read/2","type":"callback"},{"doc":"Invoked to handle configuration changes. Other internal command types may be added in future releases. Returning {:ok, new_state} sends the response :ok to the caller and continues the loop with new state new_state . Returning {:ok, new_state, side_effects} is similar to {:ok, new_state} except it causes the leader to execute side_effects (see side_effects/0 ).","ref":"ExRaft.StateMachine.html#c:handle_system_write/2","title":"ExRaft.StateMachine.handle_system_write/2","type":"callback"},{"doc":"Invoked to handle commands from ExRaft.write/3 calls. Called when command is replicated to the majority of servers and can be safely applied to the state. Returning {reply, new_state} sends the response reply to the caller and continues the loop with new state new_state . Returning {reply, new_state, side_effects} is similar to {reply, new_state} except it causes the leader to execute side_effects (see side_effects/0 ).","ref":"ExRaft.StateMachine.html#c:handle_write/2","title":"ExRaft.StateMachine.handle_write/2","type":"callback"},{"doc":"Invoked when the server is started. ExRaft.start_server/3 and ExRaft.start_server/2 will block until it returns. init_arg is the second argument passed to ExRaft.start_server/3 or [] when calling ExRaft.start_server/2 . Returning {:ok, state} will cause start_link/3 to return {:ok, pid} and the process to enter its loop. Returning {:stop, reason} will cause start_link/3 to return {:error, reason} and the process to exit with reason reason without entering the loop or calling terminate/2 . This callback is optional. If one is not implemented, init_arg will be passed along as state .","ref":"ExRaft.StateMachine.html#c:init/1","title":"ExRaft.StateMachine.init/1","type":"callback"},{"doc":"Invoked when the server is about to exit. It should do any cleanup required. reason is exit reason and state is the current state of the state machine. The return value is ignored. terminate/2 is called if the state machine traps exits (using Process.flag/2 ) and the parent process sends an exit signal. If reason is neither :normal , :shutdown , nor {:shutdown, term} an error is logged. For a more in-depth explanation, please read the &quot;Shutdown values (:shutdown)&quot; section in the Supervisor module and the GenServer.terminate/2 callback documentation. This callback is optional.","ref":"ExRaft.StateMachine.html#c:terminate/2","title":"ExRaft.StateMachine.terminate/2","type":"callback"},{"doc":"Invoked when the server transitions to a new raft state. The return value is ignored. This callback is optional.","ref":"ExRaft.StateMachine.html#c:transition/2","title":"ExRaft.StateMachine.transition/2","type":"callback"},{"doc":"The state machine reply.","ref":"ExRaft.StateMachine.html#t:reply/0","title":"ExRaft.StateMachine.reply/0","type":"type"},{"doc":"Side effect values.","ref":"ExRaft.StateMachine.html#t:side_effect/0","title":"ExRaft.StateMachine.side_effect/0","type":"type"},{"doc":"Side effects executed only by the leader.","ref":"ExRaft.StateMachine.html#t:side_effects/0","title":"ExRaft.StateMachine.side_effects/0","type":"type"},{"doc":"The state machine state.","ref":"ExRaft.StateMachine.html#t:state/0","title":"ExRaft.StateMachine.state/0","type":"type"},{"doc":"ExRaft An Elixir implementation of the raft consensus protocol. This library takes an in-memory approach for storing log entries and the state machine state, therefore it is not really suitable for storing large amounts of data. Since it's in-memory only though, it's supposed to be fast. To get started, check out the ExRaft.StateMachine documentation . To understand the internals of the project, check out the main ExRaft module documentation . For a full overview of the project, check out the full online documentation .","ref":"readme.html","title":"ExRaft","type":"extras"},{"doc":"The package can be installed by adding ex_raft to your list of dependencies in mix.exs : def deps do [ { :ex_raft , &quot;~&gt; 0.2.0&quot; } ] end","ref":"readme.html#installation","title":"ExRaft - Installation","type":"extras"},{"doc":"init_arg = [ very_cool : true ] initial_config = [ raft1 : node ( ) , raft2 : node ( ) ] # after starting these servers, they will time out and eventually elect a leader # amongst themselves { :ok , _ } = ExRaft . start_server ( YourStateMachine , init_arg , name : :raft1 , initial_config : initial_config ) { :ok , _ } = ExRaft . start_server ( YourStateMachine , init_arg , name : :raft2 , initial_config : initial_config ) # this server can never become leader unless its added to an existing cluster # since it doesn&#39;t know of any other server but is required to achieve # a minimum majority of 2 in elections and log replication { :ok , _ } = ExRaft . start_server ( YourStateMachine , init_arg , name : :raft3 , min_majority : 2 ) # we could pick any server to await the leader leader = ExRaft . await_leader ( :raft1 ) :ok = ExRaft . add_server ( leader , :raft3 ) # the success result of write/3 depends on the state machine :ok = ExRaft . write ( leader , :hello ) # making a write/3 call to a follower results in an error follower = List . delete ( initial_config , leader ) |&gt; List . first ( ) { :error , { :redirect , ^ leader } } = ExRaft . write ( follower , :hello ) # if we stop the active leader (or it crashes) # the rest of the cluster will be able to recover :ok = ExRaft . stop_server ( leader ) :ok = ExRaft . trigger_election ( :raft3 ) new_leader = ExRaft . await_leader ( :raft3 ) true = Enum . member? ( [ raft2 : node ( ) , raft3 : node ( ) ] , new_leader )","ref":"readme.html#example","title":"ExRaft - Example","type":"extras"},{"doc":"All contributions must adhere to the guidelines below: https://github.com/christopheradams/elixir_style_guide https://github.com/christopheradams/elixir_style_guide#modules https://hexdocs.pm/elixir/master/library-guidelines.html https://hexdocs.pm/elixir/master/writing-documentation.html","ref":"readme.html#contributing","title":"ExRaft - Contributing","type":"extras"},{"doc":"ExRaft source code is released under Apache License 2.0. Check LICENSE file for more information.","ref":"readme.html#license","title":"ExRaft - License","type":"extras"}]